/***************************************************************************
 *   Copyright (C) 2003-2004 by Raphael Langerhorst                        *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#ifndef GELEMENTIDH
#define GELEMENTIDH

//needed prototype for friend declarations
template<class T>
class QValueListNode;

namespace GCS
{

//will be defined in the implementation
class IDRange;

/**
  \class GElementID GElementID.h
  \brief Provides unique identification for elements
  @author Raphael Langerhorst
  
  Do NOT use ID 0 for elements, it represents "no ID".
*/

class GElementID
{  
  private:
    
    /**
      * The Range of free IDs
      * @todo as soon as the server side is split up in more processes then this must be a pointer list with many ranges as the GWE needs to add ranges.
      */
    static IDRange FreeIDs;
    
  public:
    
    static GElementID getNextFreeID();
    
    static void initFreeIDs(unsigned long low, unsigned long high);
    
  private:
  
    /**
     * The ID is represented by an unsigned long integer.
     */
    unsigned long ID;

  private:
    /**
     * For some reason QValueListNode requires a default constructor.
     * to avoid instantiation without a value everywhere else this
     * constructor is private.
     *
     * @todo This is a workaround and should be removed as soon as Qt
     * @todo does not depend on such a constructor for QValueList anymore.
     */
    GElementID()
    : ID(0)
    {}
    
  public:
  
    /**
     * Public constructor initializes to given ID.
     */
    GElementID(unsigned long ID) : ID(ID) {}

    /**
     * Copy constructor.
     */
    GElementID(const GElementID& original) : ID(original.ID) {}

    /**
     * @return copy of the ID.
     */
    unsigned long getID() const
    {
      return this->ID;
    }

    /**
     * This enables comparing of element IDs.
     * @return true when both GElementID objects have the same ID.
     */
    bool operator==(const GElementID& id) const
    {
      if (id.ID == this->ID)
        return true;
      return false;
    }
    
    /**
     * This enables comparing of element IDs.
     * @return true when both GElementID objects have different IDs.
     */
    bool operator!=(const GElementID& id) const
    {
      if (id.ID != this->ID)
        return true;
      return false;
    }
    
    //this is a workaround because QValueListNode needs
    //a default constructor;
    //
    //this constructor is declared private in GElementID
    //and made only available to this class
    //
    //note: QValueListNode is used by QValueList
    
    /**
     * Gives QValueListNode access to the private default
     * constructor - see GElementID() for details.
     * 
     */
    friend class QValueListNode<GElementID>;
};

}

#endif
