/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#ifndef GBEGREPARENTINFLUENCE_H
#define GBEGREPARENTINFLUENCE_H

#include <GElementInfluence.h>
#include <GElementID.h>
#include <GEnergy.h>
#include <GVector3.h>


namespace GBE {

/**
\class GReparentInfluence
\brief Influence with reparenting information for all affected elements.
@author Raphael Langerhorst

Elements that receive such an influence have to update their parent
information according to the new parent given by the influence.
These influences are sent by the parent of an element if this parent
detects that the child has either left this parent or entered another
child in terms of position/size. See GReparentAgent for details.

@see GReparentAgent
*/

class GReparentInfluence : public GCS::GElementInfluence
{
  protected:
    /**
     * The element that receives a new parent.
     */
    const GCS::GElementID ReparentedElement;
    
    /**
     * The ID of the new parent.
     */
    const GCS::GElementID NewParent;
    
    /**
     * Positions and rotations are relative to the current parent.
     * The reparented element should stay at the same
     * absolute position, thus it needs to modify its
     * relative position because the parent changed.
     * MoveDistance tells the reparented element how
     * much the relative position changes.
     */
    const GCS::GVector3 MoveDistance;
    
    /**
     * Positions and rotations are relative to the current parent.
     * The reparented element should stay at the same absolute
     * position and keep it's rotation, thus it needs to modify its
     * relative rotation because the parent changed.
     * Rotate tells the reparented element how
     * much the relative rotation changes.
     * @note The position of the reparented element is also rotated
     * around the new parent.
     */
    const GCS::GVector3 Rotate;
    
  public:
  
    /**
     * Constructor.
     */
    GReparentInfluence(const GCS::GElementID& source, const GCS::GEnergy& energy,
                       const GCS::GElementID& reparented_element,
                       const GCS::GElementID& new_parent,
                       const GCS::GVector3& move_distance,
                       const GCS::GVector3& rotate);

    /**
     * Virtual destructor.
     */
    virtual ~GReparentInfluence();
    
    /**
     * @returns the ID of the element that receives a new parent.
     */
    const GCS::GElementID& getReparentedElement();
    
    /**
     * @returns the ID of the new parent.
     */
    const GCS::GElementID& getNewParent();

    /**
     * @returns the relative position change in order to keep the same absolute position.
     */
    const GCS::GVector3& getMoveDistance();
    
    /**
     * @returns the relative rotation change in order to keep the same absolute position and rotation.
     */
    const GCS::GVector3& getRotate();
};

};

#endif
