/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/


#ifndef GWEINTERFACEH
#define GWEINTERFACEH

#include <GElement.h>
#include <GElementID.h>
#include <GElementInfluence.h>
#include <GVector3.h>

#include <qobject.h>
#include <qvaluelist.h>

namespace GWE
{

/**
  \class GWEInterface GWEInterface.h
  \brief Interface definition for G World Engine (GWE) implementations
  @author Raphael Langerhorst
  
  GWEInterface defines the public interface of the GWE.
  The GWEs single purpose is persistent element management and thus
  has a very slim and effective interface definition. Whether this
  implementation is a dynamic network distributed hierarchical
  server structure with database backend or a simple local implementation
  does not matter.
  
  @note Some documentation given here on this class (or interface)
  depends on the actual implementation that is used; the implementation
  is "supposed" to do what the documentation of the method say.
  
  @todo use exceptions as they provide a better way for failure description.
*/

class GWEInterface : public QObject
{
  Q_OBJECT
  public:
  
  
  // ELEMENT MANAGEMENT //
  
    /**
     * Virtual destructor to allow clean memory management
     * with derived classes.
     */
    virtual ~GWEInterface() {}
  
    /**
     * Adds given element to the GWE
     * 
     * If false is returned, either the element already exists or
     * it could not be added for various reasons.
     *
     * @return true on success
     */
    virtual bool add(GCS::GElement*) = 0;
  
    /**
     * Open an element for reading only.
     *
     * @return NULL if element does not exist.
     */
    virtual const GCS::GElement* read(const GCS::GElementID&) const = 0;
    
    /**
     * Open an element for general purposes (e.g. executing the agents,...)
     *
     * @note Every element should only be opened once.
     *
     * @return NULL if element does not exist or is already opened.
     * @see close()
     */
    virtual GCS::GElement* open(const GCS::GElementID&) = 0;
    
    /**
     * Closes given element.
     *
     * @return true when the specified element was found and closed;
     * @see open()
     */
    virtual bool close(GCS::GElement*) = 0;
    
    /**
     * In order to persistently remove an element this method needs
     * to be used, it makes sure that the element stops executing and is
     * properly removed (persistent) from memory.
     *
     * Although it may take some time until the element is truly removed,
     * the call itself is (or should be) non-blocking.
     *
     * @return true if the specified element was found and queued for deletion.
     */
    virtual bool postDelete(GCS::GElement*) = 0;
    
    
  public slots:
        
  // INFLUENCE MANAGEMENT //
  
    /**
     * Forwards given influence to all elements that are near to the
     * sender() of this signal.
     *
     * @see GElement::radiateInfluence()
     */
    virtual void radiateInfluence(GCS::GElementInfluence& influence) = 0;
    
    /**
     * Routes an influence from an element to given destination.
     *
     * @see GElement::routeInfluence()
     */
    virtual void routeInfluence(const GCS::GElementID& destination, GCS::GElementInfluence& influence) = 0;
    
    
  signals:
  
  // SIGNALS INTERESTING FOR DATA MANAGEMENT (DB,...) //
  
    /**
     * Emitted when an element is opened.
     * @note Of course emitting needs be done in the derived class.
     */
    void elementOpened(const GCS::GElement* element);
    
    /**
     * Emitted when an element is closed.
     * @note Of course emitting needs be done in the derived class.
     */
    void elementClosed(const GCS::GElement* element);
  
};

}

#endif
