/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/


#ifndef OPENGLFRAMEH
#define OPENGLFRAMEH

//GBE includes
#include <GOpenGLForm.h>

//G includes
#include <GVector3.h>
#include <GElementID.h>
#include <GElement.h>

#include <GWEInterface.h>

//Qt includes
#include <qgl.h>
#include <qmutex.h>
#include <qthread.h>
#include <qvaluelist.h>

//use G Core System classes
using namespace GCS;

//basic element classes
using namespace GBE;

//for the G world engine
using namespace GWE;

//this belongs to the client engine
namespace GCE
{

/**
 * \class CViewProperties OpenGLFrame.h
 * \brief Holds information that is relevant for camera settings
 * @author Raphael Langerhorst
 * @todo Move this class into GOpenGLFrame as it needs not really be independent.
 */
class CViewProperties
{
  public:
  
    CViewProperties()
    : ViewPosition(0,0,-1),
      ViewTarget(0,0,0),
      ViewUp(0,1,0),
      fov(90),
      NearClippingPlane(0.01),
      FarClippingPlane(10),
      width(768),
      height(1024),
      CameraElement(0)
    {
    }
  
    GVector3 ViewPosition;
    GVector3 ViewTarget;
    GVector3 ViewUp;
    double fov;
    double NearClippingPlane;
    double FarClippingPlane;
    
    int width, height;
    
    /**
     * Holds the ID of the element for which the
     * view properties apply since this modifies
     * the view transformation.
     * @note If CameraElement.getID() returns 0 then
     *       there is no parent transformation.
     */
    GElementID CameraElement;
    
    void update()
    {          
      glViewport(0,0,width,height);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(fov,(double)width/(double)height,NearClippingPlane,FarClippingPlane);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
    }
};

/**
 * \class GOpenGLFrame GOpenGLFrame.h
 * \brief A frame that can draw GOpenGLForm objects.
 * @author Raphael Langerhorst
 *
 * @todo The client interface could use MUCH improvement, its currently a very basic implementation.
 * This class represents the 3D interface. It allows for hierarchical rendering
 * of elements, specifically elements that use GOpenGLForm.
 */

class OpenGLFrame : public QGLWidget, protected QThread, public QMutex
{
  Q_OBJECT
  private:
    
    const GWEInterface* Gwe;
    
    QValueList<GElementID> TopElements;
    
    void renderChildren(const GElement* element);
    
  protected:
  
    bool stop_rendering;
  
    int sleep_time_ms;
  
    virtual void run();
    
    
  public:
    
    CViewProperties ViewProperties;

    OpenGLFrame(const GWEInterface* gwe, QWidget* parent = 0, const char* name = 0);

    ~OpenGLFrame()
    {}
    
  public slots:
    
    /**
     * starts a thread that constantly sends paint events to this OpenGL frame
     * to ensure repainting;
     * a value of 20 means a framerate of 50 fps under optimum conditions;
     */
    virtual void startRendering(int sleep_ms_between_updates_min = 20);
    
    virtual void stopRendering();
    
    virtual void initializeGL();
    
    virtual void resizeGL(int w, int h);
    
    virtual void paintGL();
    
    virtual void addTopElement(const GElementID& element);
    
    virtual void removeTopElement(const GElementID& element);

};

}

#endif
